from typing import Optional

import typer

from satcat.cli import utils
from satcat.sdk.client import Client

app = typer.Typer()
count_option = typer.Option(10, "-n", "--count")

# Screening


@app.command()
def ephemerides(count: Optional[int] = count_option, catalog_id: Optional[str] = None):
    with Client() as client:
        if catalog_id:
            resources = client.screening.list_catalog_ephemerides(
                catalog_id, count=count, sort_field='usable_time_end'
            )
            utils.cli_output_list(resources)
        else:
            resources = client.screening.list_ephemerides(count=count)
            utils.cli_output_list(resources)

@app.command()
def events(count: Optional[int] = count_option):
    with Client() as client:
        resources = client.events.list_events(count=count)
        utils.cli_output_list(resources)


@app.command()
def screenings(count: Optional[int] = count_option, show_archived: bool = False):
    with Client() as client:
        resources = client.screening.list_screenings(
            count=count, show_archived=show_archived
        )
        utils.cli_output_list(resources)


@app.command()
def catalogs(count: Optional[int] = count_option, latest: bool = False):
    with Client() as client:
        resources = client.screening.list_catalogs(count=count, latest=latest)
        utils.cli_output_list(resources)


@app.command()
def conjunctions(screening_id: Optional[str]=None, count: Optional[int] = count_option):
    with Client() as client:
        if screening_id is None:
            resources = client.events.list_conjunctions(count=count)
        else:
            resources = client.screening.list_conjunctions(
                screening_id=screening_id, count=count
            )
        utils.cli_output_list(resources)


@app.command()
def screening_primaries(screening_id: str, count: Optional[int] = count_option):
    with Client() as client:
        resources = client.screening.list_screening_primaries(
            screening_id=screening_id, count=count
        )
        utils.cli_output_list(resources)


@app.command()
def screening_secondaries(screening_id: str, count: Optional[int] = count_option):
    with Client() as client:
        resources = client.screening.list_screening_secondaries(
            screening_id=screening_id, count=count
        )
        utils.cli_output_list(resources)


@app.command()
def ephemeris_formats():
    with Client() as client:
        resources = client.screening.list_ephemeris_formats()
        utils.cli_output(resources)


@app.command()
def conjunctions_ccsds(resource_id: str, count: Optional[int] = count_option):
    with Client() as client:
        resources = client.screening.list_conjunctions_ccsds(resource_id, count)
        utils.cli_output(resources)



from datetime import datetime
from typing import TYPE_CHECKING, List, Optional

try:
    from pydantic.v1 import BaseModel, Field
except ImportError:
    from pydantic import BaseModel, Field

from satcat.sdk.models import Model

DEFAULT_PROPAGATION_DURATION_S = 86400 * 5
DEFAULT_PROPAGATION_TIMESTEP_S = 60

if TYPE_CHECKING:
    from satcat.sdk.client import Client


class PropagationConfiguration(Model):
    start_time: Optional[datetime] = Field(
        None, description="The start time of the propagation interval (UTC)."
    )
    end_time: Optional[datetime] = Field(
        None,
        description=(
            "The end time of the propagation interval (UTC). Use either this field or"
            " target_duration"
        ),
    )
    target_duration: Optional[float] = Field(
        DEFAULT_PROPAGATION_DURATION_S,
        description=(
            "The duration of the propagation interval in seconds. Use either this field"
            " or end_time"
        ),
    )
    timestep: float = Field(
        DEFAULT_PROPAGATION_TIMESTEP_S,
        description="The step size between points of the propagation in seconds",
    )
    purpose: str = "MANUAL"
    infer_interval_from_screening: bool = Field(
        True,
        description="Whether to automatically update the interval of the propagation (otherwise configured by start_time, end_time, and target_duration) to automatically match the usable time interval of the available counterpart screening data when the Propagation is attached as a primary or secondary to a Screening."
    )

    class Config:
        allow_population_by_field_name = True


class Propagation(PropagationConfiguration):
    id: str
    status: str
    created_at: datetime
    updated_at: Optional[datetime] = None
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None
    ephemeris_id: Optional[str] = None



import time
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    BinaryIO,
    Collection,
    Dict,
    List,
    Optional,
    TextIO,
    Union,
)
from urllib.parse import urljoin
import json

from satcat.sdk.propagation import models
from satcat.sdk.settings import settings
from satcat.sdk.utils import (
    PathLike,
    SortDirection,
    open_path_or_buf,
    poll_job_for_completion,
    remove_none_vals,
)

if TYPE_CHECKING:
    from satcat.sdk.client import Client

DEFAULT_SORT_DIRECTION = SortDirection.DESC
DEFAULT_SORT_FIELD = "created_at"
BEST_CATALOG_TYPE = "SP"


class PropagationClient:
    client: "Client"

    def __init__(self, client: "Client"):
        self.client = client

    def create_propagation(
        self,
        config: Optional[models.PropagationConfiguration] = None,
        opm_file: Optional[Union[PathLike, BinaryIO]] = None,
        submit: bool = False,
        norad_id: Optional[int] = None,
    ) -> models.Propagation:
        """Create a propagation on the server.

        :param config: The configuration with which to run the propagation.
        :type config: Optional[models.PropagationConfiguration], optional
        :param opm_file: The content of a CCSDS OPM (Orbit Parameter Message)
            file to use in configuring the propagation.
        :type opm_file: Optional[TextIO], optional
        :param submit: Whether to submit the propagation instantly. If ``False``,
            the propagation may be submitted later using
            ``client.propagation.submit_propagation(propagation)``.
        :type submit: bool, optional
        :return: _description_
        :rtype: models.Propagation
        """
        if config is None:
            config = models.PropagationConfiguration()
        data = {"propagation": json.loads(config.json()), "norad_id": norad_id}
        data = remove_none_vals(data)
        path = urljoin(settings.satcat_rest_api_url, "propagations")
        params = {"submit": submit}
        res = self.client.request_authenticated(
            path, json=data, method="POST", params=params
        )
        propagation = models.Propagation.parse_obj(res.json())

        if opm_file is not None:
            propagation = self.add_opm_to_propagation(propagation, opm_file)

        if submit:
            propagation = self.submit_propagation(propagation)

        return propagation

    def get_propagation(self, id: str) -> models.Propagation:
        """Get a propagation by ID from the server.

        :param id: The ID of the resource to retrieve.
        :type id: str
        :return: The retrieved propagation.
        :rtype: models.Propagation
        """
        path = urljoin(settings.satcat_rest_api_url, f"propagations/{id}")
        res = self.client.request_authenticated(path)
        data = res.json()
        data["client"] = self.client
        return models.Propagation.parse_obj(data)

    def add_opm_to_propagation(
        self, propagation: models.Propagation, opm_file: Union[PathLike, BinaryIO]
    ) -> models.Propagation:
        """Configure a propagation using a CCSDS OPM (Orbit Parameter Message).

            Configures the initial state and maneuvers of the propagation using the data
            provided in the OPM.

            If the propagation is configured to use a ``target_duration_s`` instead of
            an explicit ``end_time`` , the propagation's ``start_time`` will be updated
            to match the epoch of the state vector in the OPM.

        :param propagation: The ``Propagation`` to update.
        :type propagation: models.Propagation
        :param opm_file: _description_
        :type opm_file: TextIO
        """
        propagation_id = propagation.id
        with open_path_or_buf(opm_file) as opm_buf:
            filename = None
            if isinstance(opm_file, Path):
                filename = opm_file.name
            if isinstance(opm_file, str):
                filename = opm_file
            if filename is None:
                filename = "USER_UPLOADED_OPM.txt"
                try:
                    filename = opm_buf.name
                except AttributeError:
                    pass  # use default name if buffer object has no name attr
            req_data = {"opm_file": (filename, opm_buf)}
            path = path = urljoin(
                settings.satcat_rest_api_url, f"propagations/{propagation_id}/opm"
            )
            res = self.client.request_authenticated(path, files=req_data, method="POST")
        return models.Propagation.parse_obj(res.json())

    def submit_propagation(self, propagation: models.Propagation) -> models.Propagation:
        """Submit the Propagation for asynchronous processing on the server.
            Note that propagations may only be submitted if they have the ``"CREATED"``
            status.

        :param propagation: The Propagation to submit.
        :type propagation: models.Propagation
        :return: The submitted Propagation.
        :rtype: models.Propagation
        """

        propagation_id = propagation.id
        path = path = urljoin(
            settings.satcat_rest_api_url, f"propagations/{propagation_id}/submit"
        )
        res = self.client.request_authenticated(path, method="PUT")
        return models.Propagation.parse_obj(res.json())

    def await_propagation_completion(
        self,
        propagation: models.Propagation,
        poll_interval: int = 5,
        timeout: int = 3600,
    ) -> models.Propagation:
        """Synchronously await the completion of a propagation.

        :param propagation: The propagation to await.
        :type propagation: models.Propagation
        :param poll_interval: Interval in seconds with which to poll for results.
            Defaults to 5.
        :type poll_interval: int, optional
        :param timeout: Length in seconds after which the request should time out.
            ``TimeoutError`` is raised if the timeout expires before the propagation
            is completed. Defaults to 3600.
        :type timeout: int, optional
        :return: The completed Propagation.
        :rtype: models.Propagation
        """
        return poll_job_for_completion(
            lambda: self.get_propagation(propagation.id),
            poll_interval=poll_interval,
            timeout=timeout,
        )



from .models import (
    RSO,
    Catalog,
    Conjunction,
    Ephemeris,
    EphemerisType,
    HardBodyShape,
    Screenable,
    Screening,
    ScreeningConfiguration,
    ScreeningType,
    GeminiScreening,
)



import json
import os
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Optional, Literal
from urllib.parse import urljoin, urlparse, urlunparse

try:
    from pydantic.v1 import AnyUrl, BaseModel, BaseSettings, Extra, Field, validator
except ImportError:
    from pydantic import AnyUrl, BaseModel, BaseSettings, Extra, Field, validator


class OutputFormat(str, Enum):
    JSON = "json"
    RICH = "rich"


CONFIG_PATH = None


def json_config_settings_source(settings: BaseSettings) -> Dict[str, Any]:
    """
    A simple settings source that loads variables from a JSON file
    at the project's root.

    Here we happen to choose to use the `env_file_encoding` from Config
    when reading `config.json`
    """
    try:
        if CONFIG_PATH is not None:
            env_path = CONFIG_PATH
        else:
            env_path = os.getenv("SATCAT_CONFIG_PATH", "config.json")
        encoding = settings.__config__.env_file_encoding
        loaded_json = json.loads(Path(env_path).read_text(encoding))
        lowered_json = {key.lower(): val for key, val in loaded_json.items()}
        return lowered_json
    except FileNotFoundError:
        return {}


class CLISettings(BaseModel):
    is_cli: bool = False
    output_format: OutputFormat = OutputFormat.RICH
    show_progress: bool = True


class Settings(BaseSettings):
    auth_method: Literal['password', 'client_credentials'] = "client_credentials"
    auth_username: Optional[str] = None
    auth_password: Optional[str] = None
    auth_client_id: Optional[str] = None
    auth_client_secret: Optional[str] = None
    auth_url: AnyUrl = "https://api.satcat.com/oauth/token/"
    satcat_base_url: AnyUrl = "https://satcat.com/"
    satcat_rest_api_url_override: Optional[AnyUrl] = None
    satcat_web_api_url_override: Optional[AnyUrl] = None
    default_timeout: int = 1200
    cli: CLISettings = Field(default_factory=CLISettings)

    @validator("satcat_base_url")
    def satcat_base_url_must_end_with_slash(cls, v: str):
        if not v.endswith("/"):
            v = v + "/"
        return v

    class Config:
        env_file_encoding = "utf-8"
        extra = Extra.allow
        env_prefix = "satcat_"

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                json_config_settings_source,
                env_settings,
                file_secret_settings,
            )

    @property
    def satcat_rest_api_url(self) -> str:
        if self.satcat_rest_api_url_override is not None:
            return self.satcat_rest_api_url_override
        else:
            parsed = urlparse(self.satcat_base_url)
            netloc_parts = parsed.netloc.split('.')
            netloc_parts.insert(0, "api")
            new_netloc = '.'.join(netloc_parts)
            new_url = urlunparse(parsed._replace(netloc=new_netloc))
            return new_url

    @property
    def satcat_web_api_url(self) -> str:
        if self.satcat_web_api_url_override is not None:
            return self.satcat_web_api_url_override
        else:
            return urljoin(self.satcat_base_url, "/api")

settings = Settings()


def load_settings_file(config_path: Path):
    global CONFIG_PATH
    CONFIG_PATH = config_path

    new_settings = Settings()
    for field, value in new_settings.__dict__.items():
        setattr(settings, field, value)



from pathlib import Path

import typer

from satcat.cli.utils import cli_output_model
from satcat.sdk.client import Client

app = typer.Typer()


@app.command()
def ephemeris(
    resource_id: str,
    file_format: str = typer.Option(
        "OEM", "--file-format", "-f", help="One of OEM, NASA, HDF5, or ORIGINAL"
    ),
    output_file: Path = typer.Option(None, "--output-file", "-o"),
):
    """Download an Ephemeris by ID."""
    with Client() as client:
        file_content = client.screening.download_ephemeris(resource_id, file_format)
    if output_file is None:
        typer.echo(file_content.read())
    else:
        with open(output_file, "w+") as outf:
            outf.write(file_content.read())


@app.command()
def plan(
    scenario_id: str,
    tradespace_id: str,
    plan_id: str,
    file_name: str,
    file_format: str = typer.Option("opm", "--file-format", "-f"),
    output_file: Path = typer.Option(None, "--output-file", "-o"),
):
    """Download an Avoidance Maneuver Plan by ID."""
    with Client() as client:
        file_content = client.avoidance.download_maneuver_plan(
            scenario_id, tradespace_id, plan_id, file_format, file_name
        )
    if output_file is None:
        typer.echo(file_content.read())
    else:
        with open(output_file, "w+") as outf:
            outf.write(file_content.read())



import json
from typing import List, Optional

import typer
from typer import Abort

from satcat.cli.utils import (
    cli_output,
    cli_output_model,
    validate_model_from_input,
)

from satcat.sdk.client import Client

app = typer.Typer()




import typer

from satcat.cli.utils import cli_output, cli_output_model
from satcat.sdk.client import Client

app = typer.Typer()


# Screening


@app.command()
def ephemeris(resource_id: str):
    """Retrieve an Ephemeris by ID."""
    with Client() as client:
        resource = client.screening.get_ephemeris(resource_id)
        cli_output_model(resource)


@app.command()
def screening(resource_id: str):
    """Retrieve a Screening by ID."""
    with Client() as client:
        resource = client.screening.get_screening(resource_id)
        cli_output_model(resource)


@app.command()
def catalog(resource_id: str):
    """Retrieve a Catalog by ID."""
    with Client() as client:
        resource = client.screening.get_catalog(resource_id)
        cli_output_model(resource)


@app.command()
def conjunction_ccsds(screening_id: str, conjunction_id: str):
    """Retrieve a single CCSDS formatted Conjunction by Screening ID."""
    with Client() as client:
        resource = client.screening.get_conjunction_ccsds(screening_id, conjunction_id)
        cli_output(resource)


# Propagation


@app.command()
def propagation(resource_id: str):
    """Retrieve a Propagation by ID."""
    with Client() as client:
        resource = client.propagation.get_propagation(resource_id)
        cli_output_model(resource)




import json as jsonlib
from typing import Any, Collection, Dict, List, Literal, Optional, Union
from urllib.error import HTTPError
from urllib.parse import urljoin

import requests

from satcat.cli import utils as cli_utils
from satcat.sdk.events import models as event_models
from satcat.sdk.events.event_client import EventClient
from satcat.sdk.propagation import models as propagation_models
from satcat.sdk.propagation.propagation_client import PropagationClient
from satcat.sdk.screening import models as screening_models
from satcat.sdk.screening.screening_client import ScreeningClient
from satcat.sdk.settings import settings
from satcat.sdk.utils import SortDirection


class Client:
    """Base Client class which may be used to make requests to Satcat APIs.
    May be used as a context manager with the ``with`` keyword:

    .. code-block:: python

        with Client() as client:
            client.request(...)

    If not used with the ``with`` keyword, it is recommended to explicitly close
    the Client with ``client.close()`` after the application is finished making requests.
    """

    session: requests.Session

    def __init__(self, default_page_size: int = 32):
        """
        :param default_page_size: Default page size to use when sending repeated listing
            requests to a collection endpoint, defaults to 10
        :type page_size: int, optional
        """
        self.session = requests.Session()
        self.access_token = None
        self.authenticated = False
        self.default_page_size = default_page_size

    def __enter__(self):
        return self

    def __exit__(self, *exc_args):
        self.close()

    def close(self):
        """Close the client and any associated network connections."""
        self.session.close()

    def authenticate_session(self):
        """Authenticate the HTTP session with the authentication provider using the
            authentication method determined by ``satcat.sdk.settings.auth_method`` .

        This is done automatically when using ``request_authenticated`` .
        """
        if settings.auth_method == "password":
            path = urljoin(settings.satcat_rest_api_url, "login")
            res = self.request(
                path,
                method="POST",
                data={
                    "username": settings.auth_username,
                    "password": settings.auth_password,
                    "grant_type": "password",
                },
            )
            try:
                token = res.json()["access_token"]
            except Exception:
                raise ValueError(
                    f"Authentication Error ({res.status_code}): {res.text}"
                ) from None
            self.access_token = token
            self.session.headers["Authorization"] = f"Bearer {token}"
        elif settings.auth_method == "client_credentials":
            payload = f"grant_type=client_credentials&client_id={settings.auth_client_id}&client_secret={settings.auth_client_secret}"
            headers = {"content-type": "application/x-www-form-urlencoded"}
            response = requests.post(
                urljoin(settings.satcat_rest_api_url, "/oauth/token/"), data=payload, headers=headers, timeout=3.0
            )

            if response.status_code == 200:
                token = response.json()["access_token"]
            else:
                raise ValueError("Failed to authenticate with client_credentials flow using provided credentials")
            self.session.headers["Authorization"] = f"Bearer {token}"
            self.access_token = token
        self.authenticated = True

    def request(
        self,
        path: str,
        method: str = "GET",
        data: Optional[dict] = None,
        json: Optional[dict] = None,
        params: Optional[dict] = None,
        files=None,
        api: Literal['rest', 'web'] = 'rest'
    ) -> requests.Response:
        """Make an HTTP request to a Satcat API.
        .. warning:: For most use cases, this function should only be called internally by
        high level functions inside the SDK, such as those inside ``client.screening`` .
        """
        if api == 'rest':
            url = urljoin(settings.satcat_rest_api_url, path)
        elif api == 'web':
            url = urljoin(settings.satcat_web_api_url, path)
        else:
            raise ValueError(f"Invalid api setting {api}")

        with cli_utils.FetchingSpinner():
            response = self.session.request(
                method,
                url,
                json=json,
                data=data,
                params=params,
                files=files,
                timeout=settings.default_timeout,
            )
        return response

    def request_authenticated(
        self,
        path: str,
        method: str = "GET",
        data: Optional[dict] = None,
        json: Optional[dict] = None,
        params: Optional[dict] = None,
        ignore_failure: bool = False,
        files=None,
        api: Literal['rest', 'web'] = 'rest'
    ) -> requests.Response:
        """Ensure that the HTTP session is authorized and make an HTTP request to
            a Satcat API.

        .. warning:: For most use cases, this function should only be called internally by
            high level functions inside the SDK, such as those inside ``client.screening`` .
        """
        if not self.authenticated:
            with cli_utils.FetchingSpinner("Authenticating..."):
                self.authenticate_session()
        res = self.request(
            path, method, data=data, json=json, params=params, files=files, api=api
        )
        if not str(res.status_code).startswith("2") and not ignore_failure:
            raise HTTPError(path, res.status_code, res.text, res.headers, None)
        return res

    def request_list(
        self,
        path: str,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        params: Optional[dict] = None,
        sort_direction: Union[SortDirection, str] = SortDirection.DESC,
        sort_field: str = "created_at",
        fields: Optional[List[str]] = None
    ) -> Collection[Any]:
        """Request a list of resources from a Satcat API and automatically page through
            the results if necessary.

        .. warning:: For most use cases, this function should only be called internally by
            high level functions inside the SDK, such as those inside ``client.screening`` .
        """
        if params is None:
            params = {}
        params["sort_field"] = sort_field
        if fields is not None:
            params['fields'] = ','.join(fields)
        if isinstance(sort_direction, str):
            sort_direction = SortDirection(sort_direction)
        params["sort_direction"] = sort_direction.value.lower()
        result_items = []
        if filters is not None:
            params["filters"] = jsonlib.dumps(filters)
        if count is not None:
            params["count"] = count
            result_items = self.request_authenticated(
                path, "GET", params=params
            ).json()["items"]
        else:
            params["count"] = self.default_page_size
            result_items = []
            offset = 0
            total_count = None
            while total_count is None or offset <= total_count:
                params["offset"] = offset
                res = self.request_authenticated(path, "GET", params=params)
                data = res.json()
                items = data["items"]
                result_items.extend(items)
                offset += self.default_page_size
                total_count = data["total_count"]

        for item in result_items:
            item["client"] = self

        return result_items

    @property
    def propagation(self) -> "PropagationClient":
        """Library containing SDK functions for interacting with the
        Satcat Propagation API."""
        return PropagationClient(self)

    @property
    def screening(self) -> "ScreeningClient":
        """Library containing SDK functions for interacting with the
        Satcat Screening API."""
        return ScreeningClient(self)

    @property
    def events(self) -> "EventClient":
        """Library containing SDK functions for interacting with the
        Satcat Events API."""
        return EventClient(self)




from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, List, Optional
from uuid import UUID

from satcat.sdk.models import Model
from satcat.sdk.screening.models import RSO


class RSOMetadataModel(Model):
    country: Optional[str] = None

class RSODetailed(RSO):
    id: str
    interational_id: Optional[str] = None
    rso_metadata: Optional[RSOMetadataModel] = None

class SpacecraftConjunctionState(Model):
    comment_screening_data_source: Optional[str] = None
    ephemeris_name: Optional[str] = None
    covariance_method: Optional[str] = None
    ref_frame: Optional[str] = None
    intrack_thrust: Optional[str] = None
    comment_covariance_scale_factor: Optional[str] = None
    comment_exclusion_volume_radius: Optional[str] = None
    hbr_m: Optional[str] = None
    time_lastob_start: Optional[datetime] = None
    time_lastob_end: Optional[datetime] = None
    x_km: Optional[float] = None
    y_km: Optional[float] = None
    z_km: Optional[float] = None
    x_dot_km: Optional[float] = None
    y_dot_km: Optional[float] = None
    z_dot_km: Optional[float] = None
    cr_r: Optional[float] = None
    ct_r: Optional[float] = None
    ct_t: Optional[float] = None
    cn_r: Optional[float] = None
    cn_t: Optional[float] = None
    cn_n: Optional[float] = None
    crdot_r: Optional[float] = None
    crdot_t: Optional[float] = None
    crdot_n: Optional[float] = None
    crdot_rdot: Optional[float] = None
    ctdot_r: Optional[float] = None
    ctdot_t: Optional[float] = None
    ctdot_n: Optional[float] = None
    ctdot_rdot: Optional[float] = None
    ctdot_tdot: Optional[float] = None
    cndot_r: Optional[float] = None
    cndot_t: Optional[float] = None
    cndot_n: Optional[float] = None
    cndot_rdot: Optional[float] = None
    cndot_tdot: Optional[float] = None
    cndot_ndot: Optional[float] = None
    cdrg_r: Optional[float] = None
    cdrg_t: Optional[float] = None
    cdrg_n: Optional[float] = None
    cdrg_rdot: Optional[float] = None
    cdrg_tdot: Optional[float] = None
    cdrg_ndot: Optional[float] = None
    cdrg_drg: Optional[float] = None
    csrp_r: Optional[float] = None
    csrp_t: Optional[float] = None
    csrp_n: Optional[float] = None
    csrp_rdot: Optional[float] = None
    csrp_tdot: Optional[float] = None
    csrp_ndot: Optional[float] = None
    csrp_drg: Optional[float] = None
    csrp_srp: Optional[float] = None
    ephemeris_id: Optional[UUID] = None
    screening_id: Optional[UUID] = None
    altitude_km: Optional[float] = None



class ConjunctionMinimal(Model):
    cdm_id: Optional[int] = None
    tca: datetime
    miss_distance_km: float
    event_id: str
    originator: Optional[str] = None

    collision_probability: Optional[float] = None
    collision_probability_method: Optional[str] = None
    original_collision_probability: Optional[float] = None
    original_collision_probability_method: Optional[str] = None
    primary: Optional[RSO] = None
    secondary: Optional[RSO] = None
    creation_date: Optional[datetime] = None
    key: Optional[str] = None

class ConjunctionDetailed(ConjunctionMinimal):
    id: str
    event_id: str
    created_at: datetime
    updated_at: datetime
    designation: Optional[str] = None
    key_number: Optional[int] = None
    relative_velocity_mag_km_s: Optional[float] = None
    relative_position_r_km: Optional[float] = None
    relative_position_i_km: Optional[float] = None
    relative_position_c_km: Optional[float] = None
    relative_velocity_r_km_s: Optional[float] = None
    relative_velocity_i_km_s: Optional[float] = None
    relative_velocity_c_km_s: Optional[float] = None
    frisbee_collision_probability: Optional[float] = None
    frisbee_collision_probability_method: Optional[str] = None
    original_cdm_pc: Optional[float] = None
    angle_of_approach: Optional[float] = None
    mahalanobis_distance: Optional[float] = None
    relative_miss_sig_m: Optional[float] = None
    relative_sig_r_u_m: Optional[float] = None
    relative_sig_it_u_m: Optional[float] = None
    relative_sig_ct_u_m: Optional[float] = None
    is_intra_org: Optional[bool] = None

    primary : Optional[RSODetailed] = None
    secondary : Optional[RSODetailed] = None

    primary_state: Optional[SpacecraftConjunctionState] = None
    secondary_state: Optional[SpacecraftConjunctionState] = None



class Event(Model):
    id: str
    key: str
    created_at: datetime
    updated_at: datetime
    tca_range_start: datetime
    tca_range_end: datetime
    default_cdm: Optional[ConjunctionMinimal]=None




from urllib.parse import urljoin

import typer

from satcat import __app_name__, __version__
from satcat.cli.create import app as create_app
from satcat.cli.delete import app as delete_app
from satcat.cli.download import app as download_app
from satcat.cli.get import app as get_app
from satcat.cli.list import app as list_app
from satcat.cli.submit import app as submit_app
from satcat.cli.update import app as update_app
from satcat.cli.wait import app as await_app
from satcat.sdk.client import Client
from satcat.sdk.settings import OutputFormat, settings

VERSION = __version__
SERVICE_NAME = __app_name__

app = typer.Typer()
app.add_typer(await_app, name="await")
app.add_typer(create_app, name="create")
app.add_typer(delete_app, name="delete")
app.add_typer(download_app, name="download")
app.add_typer(get_app, name="get")
app.add_typer(list_app, name="list")
app.add_typer(submit_app, name="submit")
app.add_typer(update_app, name="update")


@app.command()
def version():
    typer.echo(f"{SERVICE_NAME} version {VERSION}")


@app.command()
def health_api():
    with Client() as client:
        res = client.request(urljoin(settings.satcat_rest_api_url, "health"))
    typer.echo(res.text)


@app.callback()
def main(
    output: OutputFormat = typer.Option(None, "-o", "--output"),
    no_progress: bool = typer.Option(False, "--no-progress"),
):
    settings.cli.is_cli = True
    if output is not None:
        settings.cli.output_format = output

    settings.cli.show_progress = not no_progress


if __name__ == "__main__":
    app()



import re
import time

from datetime import datetime, timezone
from enum import Enum
from pathlib import Path
from typing import IO, Callable, Dict, Union


class SortDirection(Enum):
    ASC = "asc"
    DESC = "desc"


PathLike = Union[str, Path]


class open_path_or_buf:
    path_or_buf: Union[Path, IO]
    buf: IO
    need_to_close_buf: bool = False

    def __init__(self, path_or_buf: Union[PathLike, IO], mode: str = "r") -> IO:
        self.path_or_buf = path_or_buf
        self.mode = mode

    def __enter__(self):
        """[summary]

        :param path_or_buf: An open file or buffer or path to a file to open.
            If a buffer is passed, it must have been opened with the correct
            mode.
        :type path_or_buf: Union[Path, IO]
        :param mode: The mode with which to open the file, defaults to "r"
        :type mode: str
        :return: [description]
        :rtype: IO
        """
        if isinstance(self.path_or_buf, Path) or isinstance(self.path_or_buf, str):
            self.buf = open(self.path_or_buf, self.mode)
            self.need_to_close_buf = True
        else:
            self.buf = self.path_or_buf
            self.need_to_close_buf = False

        return self.buf

    def __exit__(self, type, value, traceback):
        if self.need_to_close_buf:
            self.buf.close()


def remove_none_vals(d: Dict) -> Dict:
    return {k: v for k, v in d.items() if v is not None}


def poll_job_for_completion(
    get_job_fn: Callable, poll_interval: int = 5, timeout: int = 3600
):
    tst = time.time()

    elapsed = 0
    job = get_job_fn()
    while job.status in ["RUNNING", "PENDING", "CREATED"] and elapsed < timeout:
        job = get_job_fn()
        elapsed = time.time() - tst
        time.sleep(poll_interval)

    if elapsed > timeout and job.status == "RUNNING":
        raise TimeoutError()

    return job


def poll_job_for_completion_percent_generator(
    get_job_fn: Callable, poll_interval: int = 5, timeout: int = 3600
):
    tst = time.time()

    elapsed = 0
    job = get_job_fn()
    while job.status in ["RUNNING", "PENDING", "CREATED"] and elapsed < timeout:
        job = get_job_fn()
        elapsed = time.time() - tst
        time.sleep(poll_interval)
        try:
            yield job.percent_complete
        except AttributeError:
            yield 0

    if elapsed > timeout and job.status == "RUNNING":
        raise TimeoutError()

    return job


def utc(timestamp: str) -> datetime:
    """Create a datetime object from an ISO8601 timestamp string.
    The timestamp string must include the "Z" suffix indicating that
    the timestamp is expressed in the UTC time system.

    It is recommended that you use this function to pass any `datetime`
    parameters to Satcat SDK functions to ensure that your `datetime`
    is created with the correct timezone information.

    :param timestamp: The timestamp string to convert
    :type timestamp: str
    :return: The created datetime object
    :rtype: datetime
    """
    if not timestamp.endswith("Z"):
        raise ValueError("UTC Timestamps must use the 'Z' UTC indicator.")
    return datetime.fromisoformat(timestamp[:-1]).replace(tzinfo=timezone.utc)


def kebab_case(text: str) -> str:
    return re.sub(r"[^a-z0-9]+", "-", text.lower()).strip("-")



import typer
from rich.progress import Progress

from satcat.cli.utils import cli_output_model, screening_await_cli
from satcat.sdk.client import Client
from satcat.sdk.settings import OutputFormat, settings

app = typer.Typer()


@app.command()
def screening(resource_id: str, timeout: int = 3600, poll_interval: int = 5):
    """Await completion of a Screening."""
    with Client() as client:
        screening = client.screening.get_screening(resource_id)
        screening = screening_await_cli(client, screening)
        cli_output_model(screening)


@app.command()
def propagation(resource_id: str, timeout: int = 3600, poll_interval: int = 5):
    """Await completion of a Propagation."""
    with Client() as client:
        propagation = client.propagation.get_propagation(resource_id)
        resource = client.propagation.await_propagation_completion(
            propagation, timeout=timeout, poll_interval=poll_interval
        )
        cli_output_model(resource)



import typer
from rich.progress import Progress

from satcat.cli.utils import cli_output_model, screening_await_cli
from satcat.sdk.client import Client
from satcat.sdk.settings import OutputFormat, settings

app = typer.Typer()


@app.command()
def screening(
    resource_id: str,
    wait: bool = typer.Option(
        False,
        "--wait",
        "-w",
        help=(
            "Whether to wait for the screening to complete or return after creating and"
            " submitting the asynchronous screening job."
        ),
    ),
):
    """Submit a Screening by ID."""

    with Client() as client:
        screening = client.screening.get_screening(resource_id)
        screening = client.screening.submit_screening(screening)
        screening = screening_await_cli(client, screening)
        cli_output_model(screening)


@app.command()
def propagation(
    resource_id: str,
    wait: bool = typer.Option(
        False,
        "--wait",
        "-w",
        help=(
            "Whether to wait for the screening to complete or return after creating and"
            " submitting the asynchronous screening job."
        ),
    ),
):
    """Submit a Propagation by ID."""
    with Client() as client:
        propagation = client.propagation.get_propagation(resource_id)
        resource = client.propagation.submit_propagation(propagation)
        if wait:
            resource = client.propagation.await_propagation_completion(propagation)
        cli_output_model(resource)



from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import List, Optional

import shutil
import typer

from satcat.cli.utils import (
    cli_output_model,
    screening_await_cli,
    validate_model_from_input,
)

from satcat.sdk.client import Client
from satcat.sdk.propagation import models as prop_models
from satcat.sdk.screening import models as screening_models

app = typer.Typer()

# Screening


@app.command()
def ephemeris(
    file: Path,
    file_format: str = typer.Argument(...),
    norad_id: Optional[int] = None,
    comments: Optional[str] = None,
    hbr_m: Optional[float] = None,
    context: Optional[screening_models.EphemerisContext] = None,
    designation: Optional[screening_models.EphemerisDesignation] = None,
):

    with Client() as client:
        resource = client.screening.create_ephemeris(
            file,
            file_format=file_format,
            filename=file.name,
            norad_id=norad_id,
            comments=comments,
            hbr_m=hbr_m,
            context=context,
            designation=designation,
        )
        cli_output_model(resource)


@app.command()
def ephemeris_from_maneuver_plan(
    scenario_id: str = typer.Option(None, "--scenario-id", "--sid"),
    tradespace_id: str = typer.Option(None, "--tradespace-id", "--tid"),
    plan_id: str = typer.Option(None, "--plan-id", "--pid"),
    designation: Optional[screening_models.EphemerisDesignation] = typer.Option(
        None, "--designation", "--d"
    ),
):

    with Client() as client:
        resource = client.avoidance.create_ephemeris_from_maneuver_plan(
            scenario_id, tradespace_id, plan_id, designation
        )
        cli_output_model(resource)


@app.command()
def screening(
    primary_ephemeris_file: Optional[Path] = typer.Option(
        None, "--primary-ephemeris-file", "--pef"
    ),
    primary_ephemeris_format: str = typer.Option(
        "AUTOMATIC", "--primary-ephemeris-format", "--pfmt"
    ),
    primary_norad_id: Optional[int] = typer.Option(
        None, "--primary-norad-cat-id", "--pid"
    ),
    primary_ephemeris_ids: Optional[List[str]] = None,
    secondary_ephemeris_ids: Optional[List[str]] = None,
    primary_opm_file: Optional[Path] = typer.Option(None, "--opm"),
    catalog_id: Optional[str] = None,
    threshold_radius_km: float = 15.0,
    threshold_radius_active_km: Optional[float] = None,
    threshold_radius_manned_km: Optional[float] = None,
    threshold_radius_debris_km: Optional[float] = None,
    default_secondary_hbr_m: float = 5.0,
    propagation_start_time: Optional[datetime] = None,
    propagation_duration: Optional[float] = None,
    propagation_timestep: Optional[float] = None,
    launch_window_start: Optional[datetime] = None,
    launch_window_end: Optional[datetime] = None,
    launch_window_cadence_s: Optional[float] = None,
    type: screening_models.ScreeningType = screening_models.ScreeningType.ON_ORBIT,
    auto_archive: bool = False,
    include_primary_vs_primary: bool = False,
    title: Optional[str] = None,
    notes: Optional[str] = None,
    submit: bool = typer.Option(False, "--submit", "-s"),
    use_best_catalog: bool = True,
    use_ephemeris_repository: bool = False,
    wait: bool = typer.Option(
        False,
        "--wait",
        "-w",
        help=(
            "Whether to wait for the screening to complete or return after creating and"
            " submitting the asynchronous screening job."
        ),
    ),
):

    if wait and not submit:
        # Cannot wait without submitting
        submit = True

    with Client() as client:
        if primary_ephemeris_ids is None:
            primary_ephemeris_ids = []
        if secondary_ephemeris_ids is None:
            secondary_ephemeris_ids = []
        primaries = [
            screening_models.Screenable(ephemeris_id=i) for i in primary_ephemeris_ids
        ]
        secondaries = [
            screening_models.Screenable(ephemeris_id=i) for i in secondary_ephemeris_ids
        ]

        if primary_opm_file is not None:
            with open(primary_opm_file) as opm_f:
                propagation = client.propagation.create_propagation(opm_file=opm_f)
            primaries.append(screening_models.Screenable(propagation_id=propagation.id))

        catalog = None

        if use_best_catalog:
            catalog = client.screening.get_latest_catalog()
            secondaries.append(screening_models.Screenable(catalog_id=catalog.id))

        elif catalog_id is not None:
            catalog = client.screening.get_catalog(catalog_id)

        if use_ephemeris_repository:
            secondaries.append(
                screening_models.Screenable(
                    ephemeris_group=screening_models.EphemerisGroup.OPERATOR_REPOSITORY
                )
            )

        if primary_ephemeris_file:
            primary_ephemeris = client.screening.create_ephemeris(
                primary_ephemeris_file,
                file_format=primary_ephemeris_format,
                norad_id=primary_norad_id,
            )
            primaries.append(
                screening_models.Screenable(ephemeris_id=primary_ephemeris.id)
            )
        elif primary_norad_id:
            if catalog:
                primaries.append(
                    screening_models.Screenable(
                        catalog_id=catalog.id, norad_id=primary_norad_id
                    )
                )
            else:
                raise ValueError(
                    "primary_norad_id can only be used if"
                    "primary_ephemeris_file or catalog_id is set"
                )

        if title is None:
            now = datetime.now()
            title = f"{now.strftime('%Y-J%j')} - SDK Screening"

        config = screening_models.ScreeningConfiguration(
            threshold_radius_km=threshold_radius_km,
            threshold_radius_active_km=threshold_radius_active_km,
            threshold_radius_manned_km=threshold_radius_manned_km,
            threshold_radius_debris_km=threshold_radius_debris_km,
            default_secondary_hbr_m=default_secondary_hbr_m,
            submit=submit,
            notes=notes,
            propagation_duration=propagation_duration,
            propagation_start_time=propagation_start_time,
            propagation_timestep=propagation_timestep,
            auto_archive=auto_archive,
            title=title,
            include_primary_vs_primary=include_primary_vs_primary,
            screening_type=type.upper(),
            launch_window_start=launch_window_start,
            launch_window_end=launch_window_end,
            launch_window_cadence_s=launch_window_cadence_s,
        )

        screening = client.screening.create_screening(
            config,
            primaries=primaries,
            submit=submit,
            secondaries=secondaries,
        )

        screening = client.screening.get_screening(screening.id)

        if wait:
            screening = screening_await_cli(client, screening)

        cli_output_model(screening)


@app.command()
def propagation(
    opm_file: Optional[Path] = typer.Option(None, "--opm-file", "--opm"),
    submit: bool = typer.Option(False, "--submit", "-s"),
    target_duration_s: Optional[float] = typer.Option(None, "--duration-sec", "-d"),
    timestep_s: Optional[float] = typer.Option(None, "--timestep-sec", "-t"),
    purpose: Optional[str] = typer.Option(None),
    wait: bool = typer.Option(
        False,
        "--wait",
        "-w",
        help=(
            "Whether to wait for the propagation to complete or return after creating"
            " and submitting the asynchronous propagation job."
        ),
    ),
):
    configuration = prop_models.PropagationConfiguration(
        timestep=timestep_s, target_duration=target_duration_s, purpose=purpose
    )

    opm_file_content = StringIO()
    if opm_file is not None:
        with open(opm_file) as f:
            shutil.copyfileobj(f, opm_file_content)
            opm_file_content.seek(0)
    else:
        opm_file_content = None

    with Client() as client:
        propagation = client.propagation.create_propagation(
            configuration, opm_file=opm_file_content, submit=submit
        )

        if wait:
            propagation = client.propagation.await_propagation_completion(propagation)

    cli_output_model(propagation)



import inspect
import json
import sys
import rich
import typer

from enum import Enum
from typing import TYPE_CHECKING, Callable, List, Optional, TextIO, Type, Union
try:
    from pydantic.v1 import BaseModel
except ImportError:
    from pydantic import BaseModel

from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table

from satcat.sdk.propagation import models as propagation_models
from satcat.sdk.screening import models as screening_models

if TYPE_CHECKING:
    from satcat.sdk.client import Client

from satcat.sdk.settings import OutputFormat, settings


class FetchingSpinner:
    global_enable = True

    def __init__(self, message: str = "Fetching data..."):
        self.message = message

    def __enter__(self):
        self.progress = None
        if (
            settings.cli.show_progress
            and settings.cli.is_cli
            and FetchingSpinner.global_enable
        ):
            self.progress = Progress(
                SpinnerColumn(),
                TextColumn(f"[progress.description]{self.message}"),
                transient=True,
            )
            self.progress.__enter__()
            self.progress.add_task(description=self.message, total=None)

        return self

    def __exit__(self, *exc_args):
        if self.progress is not None:
            self.progress.__exit__(*exc_args)


def screening_await_cli(client: "Client", screening: screening_models.Screening):
    if settings.cli.output_format == OutputFormat.RICH and settings.cli.show_progress:
        FetchingSpinner.global_enable = False
        gen = client.screening.await_screening_completion_percent_generator(screening)

        with Progress() as progress:
            task = progress.add_task("Running screening...", total=100)
            try:
                while True:
                    pct = next(gen)
                    progress.update(task, completed=pct)
            except StopIteration as stop:
                screening = stop.value
        FetchingSpinner.global_enable = True
    else:
        screening = client.screening.await_screening_completion(screening)
    return screening


def attribute_seek(base: BaseModel, path: str):
    try:
        for key in path.split("."):
            base = getattr(base, key)
        if isinstance(base, Enum):
            base = base.value
        return base
    except AttributeError:
        return None
    except TypeError:
        return None


def rich_hyperlink_field(path: str, url_format: str) -> str:
    class _hyperlink:
        def __call__(self, base: BaseModel):
            value = attribute_seek(base, path)
            url = url_format.format(value)
            return f"[link={url}]{value}[/link]"

        @property
        def field_name(self):
            return path

    return _hyperlink()


def resolve_field(
    base: BaseModel, operator: Union[str, Callable[[BaseModel], str]]
) -> str:
    if isinstance(operator, str):
        value = attribute_seek(base, operator)
    else:
        value = operator(base)
    if value is None:
        return "N/A"
    return str(value)


def resolve_field_name(field):
    def fmt(n):
        return n.replace("_", " ").replace(".", " ").title()

    try:
        return fmt(field)
    except AttributeError:
        return fmt(field.field_name)


def cli_output(
    d: Union[BaseModel, dict],
    out: TextIO = sys.stdout,
    indent: int = 2,
):
    if settings.cli.output_format == OutputFormat.JSON:
        rich.print(f"{json.dumps(d, indent=indent)}", file=out)
    elif settings.cli.output_format == OutputFormat.RICH:
        rich.print(d, file=out)


def cli_output_list(
    d: List[BaseModel],
    out: TextIO = sys.stdout,
    indent=2,
    fields: Optional[List[str]] = None,
):
    ds = [json.loads(i.json()) for i in d]

    if fields is None:
        try:
            fields = FIELDS[type(d[0])]
        except KeyError:
            try:
                fields = ds[0].keys()
            except IndexError:
                fields = []

    if settings.cli.output_format == OutputFormat.JSON:
        rich.print(json.dumps(ds, indent=indent), file=out)
    elif settings.cli.output_format == OutputFormat.RICH:
        console = Console(file=out)

        if len(d) == 0:
            rich.print("[bold red]No results found.[/bold red]", file=out)
        else:
            fields_pretty = [resolve_field_name(f) for f in fields]
            table = Table(*fields_pretty)
            # Formatting for UUIDs
            table.columns[0].width = 36
            for item in d:
                table.add_row(*[resolve_field(item, f) for f in fields])
            console.print(table)


def cli_output_model(
    model: BaseModel,
    out: TextIO = sys.stdout,
    indent=2,
    exclude_none=True,
    fields: Optional[List[str]] = None,
):
    if fields is None:
        try:
            fields = FIELDS[type(model)]
        except KeyError:
            try:
                fields = model.__fields__.keys()
            except IndexError:
                fields = []

    if settings.cli.output_format == OutputFormat.JSON:
        rich.print(model.json(indent=indent, exclude_none=exclude_none), file=out)
    elif settings.cli.output_format == OutputFormat.RICH:
        console = Console(file=out)

        fields_pretty = [resolve_field_name(f) for f in fields]
        table = Table(*fields_pretty)
        table.add_row(*[resolve_field(model, f) for f in fields])
        console.print(table)


def display_class_keys_and_values(cls_or_module):
    help_text = f"\n{cls_or_module.__name__} Keys and Allowed Values:\n\n"

    if inspect.isclass(cls_or_module):
        if hasattr(cls_or_module, "__annotations__"):
            for attr_name, attr_value in cls_or_module.__annotations__.items():
                default_value = getattr(cls_or_module, attr_name, None)
                if default_value:
                    help_text += (
                        f"  {attr_name}: {attr_value} (Default: {default_value})\n\n"
                    )
                else:
                    help_text += f"  {attr_name}: {attr_value}\n\n"
        else:
            help_text += "  (No type annotations found)\n\n"
    elif inspect.ismodule(cls_or_module):
        for name, obj in inspect.getmembers(cls_or_module):
            if inspect.isclass(obj):
                help_text += f"\nClass {name}:\n"
                help_text += display_class_keys_and_values(obj)
    else:
        help_text += "Not a class or module.\n"

    return help_text


def validate_model_from_input(
    ctx: typer.Context, model: Type[BaseModel], flag_prefix: str
):
    validation_data = {}

    for arg_idx, extra_arg in enumerate(ctx.args):
        if extra_arg.startswith(flag_prefix):
            key_value_pair = extra_arg[len(flag_prefix) :].replace("-", "_")

            if "=" in key_value_pair:
                key, value = key_value_pair.split("=", 1)
                validation_data[key] = value
            else:
                if arg_idx + 1 < len(ctx.args):
                    validation_data[key_value_pair] = ctx.args[arg_idx + 1]

    try:
        validated_model = model.parse_obj(validation_data)
        return validated_model
    except Exception:
        raise ValueError(f"Invalid values provided for {model}")


FIELDS = {
    screening_models.Ephemeris: [
        "id",
        "norad_id",
        "solution_time",
        "has_covariance",
        "notes",
        "hbr_m",
        "context",
        "designation",
    ],
    screening_models.Screening: [
        rich_hyperlink_field("id", settings.satcat_base_url + "screenings/{}"),
        "title",
        "status",
        "created_at",
        "conjunctions_count",
        "primaries_count",
        "secondaries_count",
        "percent_complete",
        "coverage_level",
    ],
    screening_models.Conjunction: [
        "primary.norad_id",
        "secondary.norad_id",
        "tca",
        "miss_distance_km",
        "collision_probability",
    ],
    screening_models.Screenable: [
        "ephemeris.id",
        "ephemeris.norad_id",
        "catalog_id",
        "norad_id",
        "coverage_level",
    ],
    screening_models.Catalog: [
        "id",
        "catalog_type",
        "epoch",
        "ready",
        "archived",
    ],
    propagation_models.Propagation: [
        "id",
        "status",
        "created_at",
        "start_time",
        "end_time",
        "target_duration",
        "timestep",
        "ephemeris_id",
    ],
}



from datetime import datetime

try:
    from pydantic.v1 import BaseModel
except ImportError:
    from pydantic import BaseModel

json_encoders = {
    # custom output conversion for datetime
    datetime: lambda dt: dt.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
}


class Model(BaseModel):
    class Config:
        orm_mode: bool = True
        arbitrary_types_allowed: bool = True
        fields = {
            "client": {
                "exclude": ...,
            }
        }
        json_encoders = json_encoders



import json
import os
from io import StringIO
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Collection,
    Dict,
    List,
    Optional,
    TextIO,
    Union,
)
from urllib.parse import urljoin

from satcat.sdk.propagation import models as prop_models
from satcat.sdk.screening import models
from satcat.sdk.settings import settings
from satcat.sdk.utils import (
    PathLike,
    SortDirection,
    open_path_or_buf,
    poll_job_for_completion,
    poll_job_for_completion_percent_generator,
    remove_none_vals,
)

if TYPE_CHECKING:
    from satcat.sdk.client import Client

DEFAULT_SORT_DIRECTION = SortDirection.DESC
DEFAULT_SORT_FIELD = "created_at"
BEST_CATALOG_TYPE = "SP"


ScreenableType = Union[
    models.Screenable, models.Ephemeris, models.Catalog, prop_models.Propagation, models.ExternalCatalog
]

ScreenableObject = Union[models.Ephemeris, models.Catalog, prop_models.Propagation, models.ExternalCatalog]


class ScreeningClient:
    client: "Client"

    def __init__(self, client: "Client"):
        self.client = client

    def get_ephemeris(self, id: str) -> models.Ephemeris:
        """Get an Ephemeris by ID from the server.

        :param id: The ID of the resource to retrieve.
        :type id: str
        :return: The retrieved Ephemeris.
        :rtype: models.Ephemeris
        """
        path = urljoin(settings.satcat_rest_api_url, f"ephemerides/{id}")
        res = self.client.request_authenticated(path)
        data = res.json()
        data["client"] = self.client
        return models.Ephemeris.parse_obj(data)

    def download_ephemeris(self, id: str, file_format: str) -> TextIO:
        """Download the content of an ephemeris as a serialized ephemeris file.

        :param id: The ID of the resource to retrieve.
        :type id: str
        :param file_format: The file format to download the ephemeris in.

            **Note**: Only the formats ``"OEM"`` and ``"NASA"`` are currently supported. This field is case insensitive.

        :type file_format: str
        :return: A file-like object opened in text mode containing the contents of the formatted ephemeris file.
        :rtype: TextIO
        """
        path = urljoin(settings.satcat_rest_api_url, f"ephemerides/{id}/file")
        params = {"file_format": file_format.upper()}
        res = self.client.request_authenticated(path, params=params)
        outbuf = StringIO()
        outbuf.write(res.content.decode())
        outbuf.seek(0)
        return outbuf

    def get_screening(self, id: str) -> models.Screening:
        """Get a Screening by ID from the server.

        :param id: The ID of the resource to retrieve.
        :type id: str
        :return: The retrieved Screening.
        :rtype: models.Screening
        """
        path = urljoin(settings.satcat_rest_api_url, f"screenings/{id}")
        res = self.client.request_authenticated(path)
        data = res.json()
        data["client"] = self.client
        return models.Screening.parse_obj(data)

    def get_catalog(self, id: str) -> models.Catalog:
        """Get a Catalog by ID from the server.

        :param id: The ID of the resource to retrieve.
        :type id: str
        :return: The retrieved Catalog.
        :rtype: models.Catalog
        """
        path = urljoin(settings.satcat_rest_api_url, f"catalogs/{id}")
        res = self.client.request_authenticated(path)
        data = res.json()
        data["client"] = self.client
        return models.Catalog.parse_obj(data)

    def list_ephemerides(
        self,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ) -> Collection[models.Ephemeris]:
        """List the Ephemerides to which the authenticated user has access.

        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :rtype: Collection[models.Ephemeris]
        """
        params = {
            "count": count,
            "sort_field": sort_field,
            "sort_direction": sort_direction,
        }
        path = urljoin(settings.satcat_rest_api_url, "ephemerides")
        items = self.client.request_list(
            path,
            filters=filters,
            params=params,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
            fields=models.Ephemeris.__fields__.keys()
        )
        results = [models.Ephemeris.parse_obj(item) for item in items]
        return results

    def list_screenings(
        self,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
        show_archived: bool = False,
    ) -> Collection[models.Screening]:
        """List the Screenings to which the authenticated user has access.

        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :param show_archived: Whether to include archived screenings in the results.
        :type show_archived: bool, optional
        :return: The retrieved collection of resources.
        :rtype: Collection[models.Screenings]
        """
        if not show_archived:
            if filters is None:
                filters = []
            filters.append({"field": "archived", "op": "ne", "value": True})
        path = urljoin(settings.satcat_rest_api_url, "screenings")
        items = self.client.request_list(
            path,
            filters=filters,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
            fields=models.Screening.__fields__.keys()
        )
        results = [models.Screening.parse_obj(item) for item in items]
        return results

    def list_catalogs(
        self,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        latest: bool = False,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ) -> Collection[models.Catalog]:
        """List the Catalogs to which the authenticated user has access.

        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :rtype: Collection[models.Catalog]
        """
        params = {"latest": latest}
        path = urljoin(settings.satcat_rest_api_url, "catalogs")
        items = self.client.request_list(
            path,
            filters=filters,
            params=params,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
        )
        results = [models.Catalog.parse_obj(item) for item in items]
        return results

    def list_catalog_ephemerides(
        self,
        id: str,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ) -> Collection[models.Ephemeris]:
        """List the Ephemerides which belong to a Catalog by the Catalog's ID.

        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :rtype: Collection[models.Ephemeris]
        """
        path = urljoin(settings.satcat_rest_api_url, f"catalogs/{id}/ephemerides")
        items = self.client.request_list(
            path,
            filters=filters,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
        )
        results = [models.Ephemeris.parse_obj(item) for item in items]
        return results

    def list_conjunctions(
        self,
        screening_id: str,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ) -> Collection[models.Conjunction]:
        """List the Conjunctions which were generated by a Screening.

        :param screening_id: The id of the screening.
        :type screening_id: str
        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :rtype: Collection[models.Conjunction]
        """
        path = urljoin(settings.satcat_rest_api_url, f"screenings/{screening_id}/conjunctions")
        items = self.client.request_list(
            path,
            filters=filters,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
        )
        results = [models.Conjunction.parse_obj(item) for item in items]
        return results

    def list_screening_primaries(
        self,
        screening_id: str,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ) -> Collection[models.Screenable]:
        """List the primaries attached to a screening.

        :param screening_id: The id of the screening.
        :type screening_id: str
        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :r
        type: Collection[models.Screenable]
        """
        path = urljoin(settings.satcat_rest_api_url, f"screenings/{screening_id}/primaries")
        items = self.client.request_list(
            path,
            filters=filters,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
        )
        results = [models.Screenable.parse_obj(item) for item in items]
        return results

    def list_screening_secondaries(
        self,
        screening_id: str,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ) -> Collection[models.Screenable]:
        """List the secondaries attached to a screening.

        :param screening_id: The id of the screening.
        :type screening_id: str
        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :rtype: Collection[models.Screenable]
        """
        path = urljoin(settings.satcat_rest_api_url, f"screenings/{screening_id}/secondaries")
        items = self.client.request_list(
            path,
            filters=filters,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
        )
        results = [models.Screenable.parse_obj(item) for item in items]
        return results

    def get_latest_catalog(
        self, catalog_type: str = BEST_CATALOG_TYPE
    ) -> models.Catalog:
        """Get the latest ready catalog for a given catalog type which is not archived.

        :param catalog_type: The catalog type to retrieve, defaults to BEST_CATALOG_TYPE.
        :type catalog_type: str, optional
        :return: The retrieved catalog.
        :rtype: models.Catalog
        """
        catalogs = self.list_catalogs(
            latest=True, filters=[{"field": "catalog_type", "value": catalog_type}]
        )
        if len(catalogs) > 0:
            return catalogs[0]
        else:
            raise ValueError(f"No latest catalog found for catalog type {catalog_type}")

    def add_screening_primary(
        self,
        screening: models.Screening,
        primary: Optional[ScreenableObject]=None,
        norad_id: Optional[int] = None,
        from_best_catalog=True
    ) -> models.Screenable:
        """Add a primary to a screening.

        :param screening: The screening to modify.
        :type screening: models.Screening
        :param primary: The object to add to the screening.
        :type primary: ScreenableObject
        :param norad_id: Shortcut field to add an Ephemeris from a Catalog by its
            Norad Catalog ID. This field may only be used alongside `primary` if `primary`
            is a `Catalog` to add the ephemeris for a specific object from a Catalog by
            Norad ID to the screening. This is equivalent to explicitly adding the
            `Ephemeris` from the catalog to the screening directly.

            For example, if you wanted to add object 25544's ephemeris from a given catalog
            to a screening, you might use:

            .. code-block:: python

                client.screening.add_screening_primary(screening, catalog, norad_id=25544)

        :type norad_id: Optional[int], optional
        :return: The added Screenable.
        :rtype: models.Screenable
        """
        data = {}
        if isinstance(primary, models.Screening):
            data["ephemeris_id"] = primary.id
        elif isinstance(primary, models.Catalog):
            if norad_id is not None:
                data["norad_id"] = norad_id
                data["catalog_id"] = primary.id
            else:
                raise ValueError(
                    "norad_id must be supplied if primary is a Catalog to specify"
                    " a particular object."
                )
        elif isinstance(primary, models.Propagation):
            data["propagation_id"] = primary.id
        elif isinstance(primary, models.Ephemeris):
            data["ephemeris_id"] = primary.id
        elif norad_id and from_best_catalog:
            best_catalog = self.get_latest_catalog(BEST_CATALOG_TYPE)
            data["norad_id"] = norad_id
            data["catalog_id"] = best_catalog.id
        elif isinstance(primary, models.ExternalCatalog):
            raise ValueError(f"ExternalCatalogs are not supported as screening primaries.")
        else:
            raise ValueError(f"Unknown primary type {type(primary)}.")
        path = urljoin(settings.satcat_rest_api_url, f"screenings/{screening.id}/primaries")
        res = self.client.request(path, method="POST", json=data)
        return models.Screenable.parse_obj(res.json())

    def add_screening_secondary(
        self,
        screening: models.Screening,
        secondary: ScreenableObject,
        norad_id: Optional[int] = None,
    ) -> models.Screenable:
        """Add a secondary to a screening.

        :param screening: The screening to modify.
        :type screening: models.Screening
        :param secondary: The object to add to the screening.
        :type secondary: ScreenableObject
        :param norad_id: Shortcut field to add an Ephemeris from a Catalog by its
            Norad Catalog ID. This field may only be used alongside `primary` if `primary`
            is a `Catalog` to add the ephemeris for a specific object from a Catalog by
            Norad ID to the screening. This is equivalent to explicitly adding the
            `Ephemeris` from the catalog to the screening directly.

            For example, if you wanted to add object 25544's ephemeris from a given catalog
            to a screening, you might use:

            .. code-block:: python

                client.screening.add_screening_primary(screening, catalog, norad_id=25544)

        :type norad_id: Optional[int], optional
        :return: The added Screenable.
        :rtype: models.Screenable
        """
        data = {}
        if isinstance(secondary, models.Screening):
            data["ephemeris_id"] = secondary.id
        elif isinstance(secondary, models.Catalog):
            data["catalog_id"] = secondary.id
            if norad_id is not None:
                data["norad_id"] = norad_id
        elif isinstance(secondary, models.Propagation):
            data["propagation_id"] = secondary.id
        elif isinstance(secondary, models.Ephemeris):
            data["ephemeris_id"] = secondary.id
        elif isinstance(secondary, models.ExternalCatalog):
            data["external_catalog"] = secondary.value
        else:
            raise ValueError(f"Unknown secondary type {type(secondary)}.")
        path = urljoin(settings.satcat_rest_api_url, f"screenings/{screening.id}/secondaries")
        res = self.client.request(path, method="POST", json=data)
        return models.Screenable.parse_obj(res.json())

    def create_ephemeris(
        self,
        path_or_buf: Union[PathLike, BinaryIO],
        file_format: str,
        norad_id: Optional[int] = None,
        filename: Optional[str] = None,
        comments: Optional[str] = None,
        hbr_m: Optional[float] = None,
        context: Union[models.EphemerisContext, str, None] = None,
        designation: Union[models.EphemerisDesignation, str, None] = None,
    ) -> models.Ephemeris:
        """Upload an Ephemeris file to the server.
            .. note::
                Uploading an Ephemeris does not automatically trigger it to be screened.
                For this, you can use `create_screening`.

        :param path_or_buf: Path to an ephemeris file, or alternatively an open binary
            file-like object containing the contents of an ephemeris file.
            Note that the file should be opened in binary mode, even if the ephemeris
            file format is a common plaintext format.
        :type path_or_buf: Union[PathLike, BinaryIO]
        :param file_format: The file format.
        :type file_format: str
        :param norad_id: The NORAD Catalog ID of the object which the ephemeris
            file describes.

            .. note::
                Specifying the NORAD Catalog ID helps Kayhan ensure that the Ephemeris is not screened
                against other Ephemerides from the same object. While the NORAD Catalog ID field is
                optional to support Launch Screening and early-deployment screenings, we strongly
                recommend that you specify it for all on-orbit Ephemerides.

        :type norad_id: Optional[int]
        :param filename: The filename of the ephemeris file. If ``path_or_buf`` is
            a ``Path``, the filename will be automatically determined. Defaults to None
        :type filename: Optional[str], optional
        :param comments: Optional user-defined comments to attach to the file metadata.
            Defaults to None
        :type comments: Optional[str], optional
        :param hbr_m: The hard-body radius of the object in meters
        :type hbr_m: Optional[float]
        :param context: The Pathfinder context metadata for the ephemeris. For more information, see https://app.satcat.io/docs/pathfinder/operational-data/#ephemeris-metadata.
        :type context: Union[models.EphemerisContext, str, None]
        :param designation: The Pathfinder designation metadata for the ephemeris. For more information, see https://app.satcat.io/docs/pathfinder/operational-data/#ephemeris-metadata.
        :type designation: Union[models.EphemerisDesignation, str, None]
        :return: The created Ephemeris
        :rtype: models.Ephemeris
        """
        with open_path_or_buf(path_or_buf, mode="rb") as file:
            file_format = file_format.upper()
            if isinstance(path_or_buf, Path) and filename is None:
                filename = os.path.basename(path_or_buf.name)
            if isinstance(path_or_buf, str) and filename is None:
                filename = path_or_buf
            if filename is None:
                filename = "USER_UPLOADED_EPHEMERIS.txt"
                try:
                    filename = os.path.basename(file.name)
                except AttributeError:
                    pass  # use default name if buffer object has no name attr
            hbr_str = str(hbr_m) if hbr_m is not None else None
            context = (
                models.EphemerisContext(context)
                if isinstance(context, str)
                else context
            )
            designation = (
                models.EphemerisDesignation(designation)
                if isinstance(designation, str)
                else designation
            )
            data = [
                ("norad_id", (None, norad_id)),
                ("file_format", (None, file_format)),
                ("comments", (None, comments)),
                ("ephemeris_file", (filename, file)),
                ("hbr_m", (None, hbr_str)),
                ("context", (None, context.value if context is not None else None)),
                (
                    "designation",
                    (None, designation.value if designation is not None else None),
                ),
            ]
            path = urljoin(settings.satcat_rest_api_url, "ephemerides")
            res = self.client.request_authenticated(path, files=data, method="POST")
        return models.Ephemeris.parse_obj(res.json())

    def operationalize_ephemeris(self, ephemeris_id: str) -> models.Ephemeris:
        """Operationalize an ephemeris by ID.

        :param ephemeris_id: The ID of the resource to designate as 'OPERATIONAL'.
        :type ephemeris_id: str
        :return: The updated Ephemeris.
        :rtype: models.Ephemeris
        """

        path = urljoin(
            settings.satcat_rest_api_url,
            f"ephemerides/{ephemeris_id}/operationalize",
        )
        res = self.client.request_authenticated(path, json={}, method="PUT")
        ephemeris = models.Ephemeris.parse_obj(res.json())

        return ephemeris

    def resolve_to_screenable(self, input: ScreenableType) -> models.Screenable:
        if isinstance(input, models.Screenable):
            return input
        elif isinstance(input, models.Ephemeris):
            return models.Screenable(ephemeris_id=input.id)
        elif isinstance(input, models.Catalog):
            return models.Screenable(catalog_id=input.id)
        elif isinstance(input, prop_models.Propagation):
            return models.Screenable(propagation_id=input.id)
        else:
            raise TypeError(
                f"Cannot create Screenable from input of type {type(input)}"
            )

    def create_screening(
        self,
        config: Optional[models.ScreeningConfiguration] = None,
        primaries: Optional[List[ScreenableType]] = None,
        secondaries: Optional[List[ScreenableType]] = None,
        add_best_secondary_catalog: bool = False,
        add_operational_ephemeris_repository: bool = False,
        submit: bool = False,
    ) -> models.Screening:
        """Create a Screening on the server.

        :param config: The configuration with which to run the screening.
        :type config: models.ScreeningConfiguration
        :param primaries: The list of primaries which should be screened.
        :type primaries: Optional[List[models.Screenable]], optional
        :param secondaries: The list of secondaries which should be screened against.
        :type secondaries: Optional[List[models.Screenable]], optional
        :param add_best_secondary_catalog: Whether the secondary catalog determined by
            the application as "Best" (latest updated data and highest-fidelity
            available data source) should be automatically added to the screening.
        :type add_best_secondary_catalog: bool, optional
        :param add_operational_ephemeris_repository: Whether all ephemeris with a designation of
            ``OPERATIONAL`` and with a ``usable_time_end`` in the future should be automatically
            added to the screening.
        :type add_operational_ephemeris_repository: bool, optional
        :param submit: Whether to submit the screening instantly. If ``False``,
            the screening may be submitted later using
            ``client.screening.submit_screening(screening)``.
        :type submit: bool, optional
        :return: The created Screening.
        :rtype: models.Screening
        """
        if config is None:
            config = models.ScreeningConfiguration()
        data = remove_none_vals(json.loads(config.json()))

        if primaries is None:
            primaries = []
        if secondaries is None:
            secondaries = []

        if add_best_secondary_catalog:
            best_catalog = self.get_latest_catalog(BEST_CATALOG_TYPE)
            screenable = models.Screenable(catalog_id=best_catalog.id)
            secondaries = [s for s in secondaries]
            secondaries.append(screenable)

        if add_operational_ephemeris_repository:
            screenable = models.Screenable(
                ephemeris_group=models.EphemerisGroup.OPERATOR_REPOSITORY
            )
            secondaries.append(screenable)

        if primaries is not None:
            primaries = [self.resolve_to_screenable(p) for p in primaries]
            primaries_dicts = [remove_none_vals(p.dict()) for p in primaries]
            data["primaries"] = primaries_dicts
        if secondaries is not None:
            secondaries = [self.resolve_to_screenable(s) for s in secondaries]
            secondaries_dicts = [remove_none_vals(s.dict()) for s in secondaries]
            data["secondaries"] = secondaries_dicts
        params = {"submit": submit}
        path = urljoin(settings.satcat_rest_api_url, "screenings")
        res = self.client.request_authenticated(
            path, json=data, method="POST", params=params
        )
        screening = models.Screening.parse_obj(res.json())

        return screening

    def submit_screening(self, screening: models.Screening) -> models.Screening:
        """Submit the Screening for asynchronous processing on the server.
        Note that screenings may only be submitted if they have the ``"CREATED"``
        status.

        :param screening: The Screening to submit.
        :type screening: models.Screening
        :return: The submitted Screening.
        :rtype: models.Screening
        """
        path = urljoin(settings.satcat_rest_api_url, f"screenings/{screening.id}/submit")
        res = self.client.request_authenticated(path, method="PUT")
        return models.Screening.parse_obj(res.json())

    def await_screening_completion(
        self, screening: models.Screening, poll_interval: int = 5, timeout: int = 3600
    ) -> models.Screening:
        """Synchronously await the completion of a screening.

        :param screening: The screening to await.
        :type screening: models.Screening
        :param poll_interval: Interval in seconds with which to poll for results.
            Defaults to 5.
        :type poll_interval: int, optional
        :param timeout: Length in seconds after which the request should time out.
            ``TimeoutError`` is raised if the timeout expires before the screening
            is completed. Defaults to 3600.
        :type timeout: int, optional
        :return: The completed Screening.
        :rtype: models.Screening
        """
        return poll_job_for_completion(
            lambda: self.get_screening(screening.id),
            poll_interval=poll_interval,
            timeout=timeout,
        )

    def await_screening_completion_percent_generator(
        self, screening: models.Screening, poll_interval: int = 5, timeout: int = 3600
    ) -> models.Screening:  # type: ignore
        gen = poll_job_for_completion_percent_generator(
            lambda: self.get_screening(screening.id),
            poll_interval=poll_interval,
            timeout=timeout,
        )
        while True:
            try:
                yield next(gen)
            except StopIteration as e:
                return e.value

    def get_api_spec(self) -> dict:
        """Retrieve the OpenAPI specification from the server."""
        path = urljoin(settings.satcat_rest_api_url, "openapi.json")
        res = self.client.request(path)
        return res.json()

    def list_ephemeris_formats(self) -> Collection[str]:
        """Retrieve the list of supported ephemeris file formats."""
        spec = self.get_api_spec()
        return spec["components"]["schemas"]["EphemerisReadFileFormat"]["enum"]

    def list_conjunctions_ccsds(
        self,
        screening_id: str,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ) -> List[Dict[str, Any]]:
        """List the Conjunctions in CCSDS format which belong to a Screening by the Screenings's ID.

        :param screening_id: ID of Screening job for conjunctions
        :type screening_id: str
        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :rtype: List[Dict[str, Any]]
        """
        path = urljoin(settings.satcat_rest_api_url, f"screenings/{screening_id}/results/ccsds")
        items = self.client.request_list(
            path,
            filters=filters,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
        )
        sanitized_items = [
            {key: value for key, value in item.items() if key != "client"}
            for item in items
        ]
        return sanitized_items

    def get_conjunction_ccsds(
        self,
        screening_id: str,
        conjunction_id: str,
    ) -> Dict[str, Any]:
        """List single Conjunction in CCSDS format which belong to a Screening by the Screenings's ID.

        :param screening_id: ID of Screening job for conjunctions
        :type screening_id: str
        :param conjunction_id: ID of individual Conjunction
        :type conjunction_id: str
        :return: The Conjunction resource.
        :rtype: Dict[str, Any]
        """
        path = urljoin(
            settings.satcat_rest_api_url,
            f"screenings/{screening_id}/results/ccsds/{conjunction_id}",
        )
        res = self.client.request_authenticated(path)
        return res.json()



from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, List, Optional
from uuid import UUID

from satcat.sdk.models import Model
from satcat.sdk.propagation.models import Propagation

if TYPE_CHECKING:
    from satcat.sdk.client import Client


class EphemerisType(str, Enum):
    ON_ORBIT = "ON_ORBIT"
    LAUNCH = "LAUNCH"


class ScreeningType(str, Enum):
    ON_ORBIT = "ON_ORBIT"
    LAUNCH = "LAUNCH"


class EphemerisDesignation(str, Enum):
    OPERATIONAL = "OPERATIONAL"
    PREDICTIVE = "PREDICTIVE"
    DEFINITIVE = "DEFINITIVE"
    THEORETICAL = "THEORETICAL"


class EphemerisContext(str, Enum):
    ROUTINE = "ROUTINE"
    MAINTENANCE = "MAINTENANCE"
    COLA = "COLA"
    DECAY = "DECAY"
    NOVEL = "NOVEL"


class EphemerisGroup(str, Enum):
    OPERATOR_REPOSITORY = "OPERATOR_REPOSITORY"


class CoverageLevel(str, Enum):
    NO_OVERLAP = "NO_OVERLAP"
    PARTIAL = "PARTIAL"
    FULL = "FULL"


class HardBodyShape(str, Enum):
    SQUARE = "SQUARE"
    CIRCLE = "CIRCLE"


class EphemerisSource(str, Enum):
    CATALOG = "CATALOG"
    SPACE_TRACK_PUBLIC = "SPACE_TRACK_PUBLIC"
    PATHFINDER_API = "PATHFINDER_API"
    SATCAT_API = "SATCAT_API"
    USER_SCREENER = "USER_SCREENER"
    TERTIARY_SCREENER = "TERTIARY_SCREENER"
    PATHFINDER_AUTO_THIRDPARTY = "PATHFINDER_AUTO_THIRDPARTY"
    KAYHAN_INTERNAL_PROCESS = "KAYHAN_INTERNAL_PROCESS"

class ExternalCatalog(str, Enum):
    STARLINK= "STARLINK"

class Catalog(Model):
    id: str
    catalog_type: str
    archived: bool
    ready: bool
    epoch: datetime
    filename: Optional[str] = None


class RSO(Model):
    norad_id: int
    object_name: Optional[str]
    object_type: Optional[str]


class Ephemeris(Model):
    # ephemeris_type: EphemerisType = EphemerisType.ON_ORBIT
    ephemeris_source: Optional[EphemerisSource] = None
    current_operational: Optional[bool] = None
    id: Optional[str] = None
    has_covariance: Optional[bool] = None
    apogee_km: Optional[float] = None
    perigee_km: Optional[float] = None
    archived: Optional[bool] = None
    user_id: Optional[str] = None
    # catalog: Optional[Catalog] = None
    norad_id: Optional[int] = None
    solution_time: Optional[datetime] = None
    # rso: Optional[RSO] = None
    hbr_m: Optional[float] = None
    filename: Optional[str] = None
    usable_time_start: Optional[datetime] = None
    usable_time_end: Optional[datetime] = None
    context: Optional[EphemerisContext] = None
    designation: Optional[EphemerisDesignation] = None
    comments: Optional[str] = None
    data_format: Optional[str] = None
    launch_time: Optional[datetime] = None


class Screenable(Model):
    ephemeris_id: Optional[str] = None
    catalog_id: Optional[str] = None
    propagation_id: Optional[str] = None
    ephemeris: Optional[Ephemeris] = None
    catalog: Optional[Catalog] = None
    propagation: Optional[Propagation] = None
    norad_id: Optional[int] = None
    usable_time_start: Optional[datetime] = None
    usable_time_end: Optional[datetime] = None
    coverage_level: Optional[CoverageLevel] = None
    coverage_ratio: Optional[float] = None
    ephemeris_group: Optional[EphemerisGroup] = None
    external_catalogs: Optional[List[ExternalCatalog]] = None


class ScreeningConfiguration(Model):
    threshold_radius_km: float = 15.0
    threshold_radius_active_km: Optional[float] = None
    threshold_radius_manned_km: Optional[float] = None
    threshold_radius_debris_km: Optional[float] = None
    default_secondary_hbr_m: float = 5.0
    propagation_start_time: Optional[datetime] = None
    # propagation_duration: Optional[float] = None
    # propagation_timestep: Optional[float] = None
    auto_archive: bool = False
    include_primary_vs_primary: bool = False
    screening_type: ScreeningType = ScreeningType.ON_ORBIT
    launch_window_start: Optional[datetime] = None
    launch_window_end: Optional[datetime] = None
    launch_window_cadence_s: Optional[float] = None
    title: Optional[str] = "Screening"
    notes: Optional[str] = "Created using Satcat SDK"
    hard_body_shape: HardBodyShape = HardBodyShape.SQUARE


class Conjunction(Model):
    cdm_id: Optional[int] = None
    tca: datetime
    miss_distance_km: float
    originator: Optional[str] = None

    launch_time: Optional[datetime] = None
    id: Optional[str] = None
    collision_probability: Optional[float] = None
    collision_probability_method: Optional[str] = None
    created_at: Optional[datetime] = None
    primary: Optional[Ephemeris] = None
    secondary: Optional[Ephemeris] = None
    key_number: int

    @property
    def key(self) -> str:
        originator_prefix = "X"
        if self.originator:
            if self.originator.upper() in {"JSPOC", "CSPOC"}:
                originator_prefix = "S"
            elif self.originator.upper() == "KAYHAN":
                originator_prefix = "K"
        return f"CDM-{originator_prefix}{self.key_number}"


class Screening(ScreeningConfiguration):
    id: str
    status: Optional[str] = None
    screening_type: Optional[ScreeningType] = ScreeningType.ON_ORBIT
    created_at: Optional[datetime] = None
    primaries_count: Optional[int] = None
    secondaries_count: Optional[int] = None
    conjunctions_count: Optional[int] = None
    updated_at: Optional[datetime] = None
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None
    percent_complete: Optional[float] = None
    coverage_level: Optional[CoverageLevel] = None
    coverage_ratio: Optional[float] = None
    primary_rsos_preview: Optional[List[RSO]] = None

    @property
    def primary_ids_preview(self) -> List[int]:
        return [r.norad_id for r in self.primary_rsos_preview]


class GeminiScreening(Model):
    id: str
    status: str
    conjunctions_under_thresholds_count: Optional[int] = None



from io import StringIO
from typing import TYPE_CHECKING, Dict, List, Optional
from urllib.parse import urljoin

from satcat.sdk.events import models
from satcat.sdk.settings import settings
from satcat.sdk.utils import SortDirection

DEFAULT_SORT_DIRECTION = SortDirection.DESC
DEFAULT_SORT_FIELD = "created_at"


if TYPE_CHECKING:
    from satcat.sdk.client import Client

class EventClient:
    client: "Client"

    def __init__(self, client: "Client"):
        self.client = client

    def list_events(
        self,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ):
        """
        List available events

        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :rtype: Collection[models.Event]
        """
        params = {
            "count": count,
            "sort_field": sort_field,
            "sort_direction": sort_direction,
        }
        path = urljoin(settings.satcat_rest_api_url, "events")
        items = self.client.request_list(
            path,
            filters=filters,
            params=params,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
            fields=models.Event.__fields__.keys()
        )
        return [models.Event.parse_obj(item) for item in items]

    def list_event_conjunctions(
        self,
        event_key: str,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ):
        """
        List available conjunctions for a specific event

        :param event_key: The key of the event to filter by.
        :type event_key: str
        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :rtype: Collection[models.ConjunctionMinimal]
        """
        params = {
            "count": count,
            "sort_field": sort_field,
            "sort_direction": sort_direction,
        }
        
        path = urljoin(settings.satcat_rest_api_url, f"events/{event_key}/cdms")
        
        items = self.client.request_list(
            path,
            filters=filters,
            params=params,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
        )
        
        return [models.ConjunctionMinimal.parse_obj(item) for item in items]

    def list_conjunctions(
        self,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ):
        """
        List available conjunctions

        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :rtype: Collection[models.ConjunctionMinimal]
        """
        params = {
            "count": count,
            "sort_field": sort_field,
            "sort_direction": sort_direction,
        }
        path = urljoin(settings.satcat_rest_api_url, "conjunction_messages")
        items = self.client.request_list(
            path,
            filters=filters,
            params=params,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
            fields=models.ConjunctionMinimal.__fields__.keys()
        )
        return [models.ConjunctionMinimal.parse_obj(item) for item in items]

    def list_conjunctions_ccsds(
        self,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        sort_field: str = DEFAULT_SORT_FIELD,
        sort_direction: "SortDirection" = DEFAULT_SORT_DIRECTION,
    ):
        """
        List available conjunctions in CCSDS format

        :param filters: Filters to apply to the list of resources, defaults to None.
        :type filters: Optional[List[Dict]], optional
        :param count: Maximum number of results to return. If None, return the entire
            collection. Defaults to None.
        :type count: Optional[int], optional
        :param sort_field: The field on which to sort results, defaults to DEFAULT_SORT_FIELD.
        :type sort_field: str, optional
        :param sort_direction: The direction in which to sort results, defaults to "asc".
        :type sort_direction: str, optional
        :return: The retrieved collection of resources.
        :rtype: Collection[models.ConjunctionMinimal]
        """
        params = {
            "count": count,
            "sort_field": sort_field,
            "sort_direction": sort_direction,
        }
        path = urljoin(settings.satcat_rest_api_url, "conjunction_messages/ccsds")
        items = self.client.request_list(
            path,
            filters=filters,
            params=params,
            count=count,
            sort_direction=sort_direction,
            sort_field=sort_field,
        )
        sanitized_items = [
            {key: value for key, value in item.items() if key != "client"}
            for item in items
        ]
        return sanitized_items

    def get_conjunction(self, cdm_id: str):
        """
        Get a specific conjunction by its cdm_id or key. Note that this returns a more details Conjunction description than the screening conjunction endpoint.

        :param cdm_id: the cdm_id or cdm key of the conjunction to retrieve.
        :type cdm_id: str
        :return: The retrieved conjunction.
        :rtype: models.Conjunction
        """
        path =f"conjunction_messages/{cdm_id}"
        res = self.client.request_authenticated(path)
        data=res.json()
        data["client"]=self.client
        return models.ConjunctionDetailed.parse_obj(data)

    def get_conjunction_ccsds(self, cdm_id: str):
        """
        Get a specific conjunction by its cdm_id or key in CCSDS format. Note that this returns a more details Conjunction description than the screening conjunction endpoint.

        :param cdm_id: the cdm_id or cdm key of the conjunction to retrieve.
        :type cdm_id: str
        :return: The retrieved conjunction.
        :rtype: models.Conjunction
        """
        path =f"conjunction_messages/{cdm_id}/ccsds"
        res = self.client.request_authenticated(path)
        return res.json()



    def download_conjunction(self, cdm_id: str, file_format: Optional[str] = "json"):
        """
        Downloads a specific conjunction message file by its CDM ID or key.

        :param cdm_id: The CDM ID or key of the conjunction to retrieve.
        :param updated: Flag indicating whether to retrieve the updated version of the conjunction file. Defaults to False.
        :type updated: Optional[bool]
        :param file_format: The desired file format for the conjunction file (e.g., "json"). Defaults to "json".
        :type file_format: Optional[str]
        :return: The content of the retrieved conjunction file.
        :rtype: bytes
        """
        path =f"conjunction_messages/{cdm_id}/file"
        params = {
            "file_format": file_format.upper()
        }
        res = self.client.request_authenticated(path, params=params)
        outbuf=StringIO()
        outbuf.write(res.content.decode())
        outbuf.seek(0)
        return outbuf